@@ -287,6 +287,11 @@ if dollar_ok:
     name = Pattern('<name>', r'[A-Z][\w$]*', flags=re.I)
 else:
     name = Pattern('<name>', r'[A-Z]\w*', flags=re.I)
+# file_name pattern is start of match '^' to end of match '$', either
+# match a single character that is not space '\S', or '|' a single
+# character that is not space at the start '\S' and end '\S' of the
+# match with anything '.*' inbetween.
+file_name = Pattern('<file_name>', '^(\S|\S.*\S)$', flags=re.I)
 digit = Pattern('<digit>', r'\d')
 underscore = Pattern('<underscore>', '_')
 binary_digit = Pattern('<binary-digit>', r'[01]')
@@ -404,7 +409,7 @@ defined_binary_op = Pattern(
 defined_operator = defined_unary_op | defined_binary_op | \
                    extended_intrinsic_operator
 abs_defined_operator = abs(defined_operator)
-defined_op = Pattern('<defined-op>', '[.]\s*[A-Z]+\s*[.]', flags=re.I)
+defined_op = Pattern('<defined-op>', '[.][A-Z]+[.]', flags=re.I)
 abs_defined_op = abs(defined_op)
 
 non_defined_binary_op = intrinsic_operator | logical_literal_constant
