@@ -1,4 +1,4 @@
-# Modified work Copyright (c) 2017-2018 Science and Technology
+# Modified work Copyright (c) 2017-2019 Science and Technology
 # Facilities Council
 # Original work Copyright (c) 1999-2008 Pearu Peterson
 
@@ -71,11 +71,47 @@
 
 import re
 import logging
+import six
 from fparser.common.splitline import string_replace_map
 from fparser.two import pattern_tools as pattern
 from fparser.common.readfortran import FortranReaderBase
 
 
+# A list of supported extensions to the standard(s)
+
+# An X edit descriptor in a format statement specifies the position
+# (forward from the current position) at which the next character will
+# be transmitted to or from a record. In standard Fortran2003 the X
+# edit descriptor must be preceeded by an integer which specifies how
+# far forward from the current position. The 'x-format' extension
+# allows the X edit descriptor to be specified without a preceeding
+# integer.
+EXTENSIONS = ["x-format"]
+
+# Cray pointers are a well known extension to the Fortran
+# standard. See http://pubs.cray.com/content/S-3901/8.6/
+# cray-fortran-reference-manual-s-3901-86/types or
+# https://gcc.gnu.org/onlinedocs/gfortran/Cray-pointers.html for
+# example. If 'cray-pointer' is specified in EXTENSIONS then this
+# extension is allowed in fparser.
+EXTENSIONS += ["cray-pointer"]
+
+# A Hollerith constant is a way of specifying a string as a sequence
+# of characters preceded by the string length and separated by an 'H'
+# e.g. 5Hhello. See
+# https://gcc.gnu.org/onlinedocs/gfortran/Hollerith-constants-support.html,
+# for example for more details. fparser currently supports Hollerith
+# constants specified in format statements when 'hollerith' is specified
+# in the EXTENSIONS list.
+EXTENSIONS += ["hollerith"]
+
+# Many compilers support the use of '$' in a fortran write statement
+# to indicate that the carriage return should be suppressed. This is
+# an extension to the Fortran standard and is supported by fparser if
+# 'dollar-descriptor' is specified in the EXTENSIONS list.
+EXTENSIONS += ["dollar-descriptor"]
+
+
 class FparserException(Exception):
     '''Base class exception for fparser. This allows an external tool to
     capture all exceptions if required.
@@ -109,7 +145,7 @@ class FortranSyntaxError(FparserException):
 
     '''
     def __init__(self, reader, info):
-        output = "at unknown location"
+        output = "at unknown location "
         if isinstance(reader, FortranReaderBase):
             output = "at line {0}\n>>>{1}\n".format(
                 reader.linecount,
@@ -132,6 +168,14 @@ class InternalError(FparserException):
         FparserException.__init__(self, new_info)
 
 
+class InternalSyntaxError(FparserException):
+    '''An exception indicating that a syntax error has been found by the
+    parser. This is used instead of `FortranSyntaxError` when the
+    reader object is not available.
+
+    '''
+
+
 def show_result(func):
     return func
 
@@ -299,7 +343,7 @@ class Base(ComparableMixin):
                 # Check all lines up to this one for content. We
                 # should be able to only check the current line but
                 # but as the line number returned is not always
-                # correct (due to coding errors) we can not assume the
+                # correct (due to coding errors) we cannot assume the
                 # line pointed to is the line where the error actually
                 # happened.
                 if string.source_lines[index].strip():
@@ -312,10 +356,10 @@ class Base(ComparableMixin):
                 # follow their lead and do not raise an exception.
                 return
             line = string.source_lines[string.linecount-1]
-            errmsg = "at line {0}\n>>>{1}\n".format(
+            errmsg = u"at line {0}\n>>>{1}\n".format(
                 string.linecount, line)
         else:
-            errmsg = "{0}: '{1}'".format(cls.__name__, string)
+            errmsg = u"{0}: '{1}'".format(cls.__name__, string)
         raise NoMatchError(errmsg)
 
     def init(self, *items):
@@ -338,7 +382,16 @@ class Base(ComparableMixin):
         return self.items
 
     def tofortran(self, tab='', isfix=None):
-        this_str = str(self)
+        '''
+        Produce the Fortran representation of this Comment.
+
+        :param str tab: characters to pre-pend to output.
+        :param bool isfix: whether or not this is fixed-format code.
+
+        :returns: Fortran representation of this comment.
+        :rtype: str
+        '''
+        this_str = six.text_type(self)
         if this_str.strip():
             return tab + this_str
         else:
@@ -365,13 +418,14 @@ content : tuple
     def match(startcls, subclasses, endcls, reader,
               match_labels=False,
               match_names=False,
-              set_unspecified_end_name=False,
               match_name_classes=(),
               enable_do_label_construct_hook=False,
               enable_if_construct_hook=False,
               enable_where_construct_hook=False,
               enable_select_type_construct_hook=False,
-              enable_case_construct_hook=False):
+              enable_case_construct_hook=False,
+              strict_order=False,
+              ):
         '''
         Checks whether the content in reader matches the given
         type of block statement (e.g. DO..END DO, IF...END IF etc.)
@@ -384,24 +438,26 @@ content : tuple
         :type reader: str or instance of :py:class:`FortranReaderBase`
         :param bool match_labels: TBD
         :param bool match_names: TBD
-        :param bool set_unspecified_end_name: TBD
         :param tuple match_name_classes: TBD
         :param bool enable_do_label_construct_hook: TBD
         :param bool enable_if_construct_hook: TBD
         :param bool enable_where_construct_hook: TBD
         :param bool enable_select_type_construct_hook: TBD
         :param bool enable_case_construct_hook: TBD
+        :param bool strict_order: Whether to enforce the order of the
+                                  given subclasses.
 
         :return: instance of startcls or None if no match is found
         :rtype: startcls
         '''
-        from fparser.two.Fortran2003 import Comment, add_comments
+        from fparser.two.Fortran2003 import Comment, Include_Stmt, \
+            add_comments_includes
         assert isinstance(reader, FortranReaderBase), repr(reader)
         content = []
 
         if startcls is not None:
-            # Deal with any preceding comments
-            add_comments(content, reader)
+            # Deal with any preceding comments and/or includes
+            add_comments_includes(content, reader)
             # Now attempt to match the start of the block
             try:
                 obj = startcls(reader)
@@ -423,8 +479,8 @@ content : tuple
             if match_names:
                 start_name = obj.get_start_name()
 
-        # A comment is always a valid sub-class
-        classes = subclasses + [Comment]
+        # Comments and Include statements are always valid sub-classes
+        classes = subclasses + [Comment, Include_Stmt]
         if endcls is not None:
             classes += [endcls]
             endcls_all = tuple([endcls]+endcls.subclasses[endcls.__name__])
@@ -484,10 +540,7 @@ content : tuple
                     start_name, end_name = content[start_idx].\
                                            get_start_name(), \
                                            content[-1].get_end_name()
-                    if set_unspecified_end_name and end_name is None and \
-                       start_name is not None:
-                        content[-1].set_name(start_name)
-                    elif end_name and not start_name:
+                    if end_name and not start_name:
                         raise FortranSyntaxError(
                             reader, "Name '{0}' has no corresponding starting "
                             "name".format(end_name))
@@ -498,8 +551,9 @@ content : tuple
                 # We've found the enclosing end statement so break out
                 found_end = True
                 break
-            # Return to start of classes list now that we've matched
-            i = 0
+            if not strict_order:
+                # Return to start of classes list now that we've matched.
+                i = 0
             if enable_if_construct_hook:
                 from fparser.two.Fortran2003 import Else_If_Stmt, Else_Stmt, \
                     End_If_Stmt
@@ -557,8 +611,6 @@ content : tuple
                             'expected <%s-name> is %s but got %s. Ignoring.'
                             % (end_stmt.get_type().lower(),
                                start_stmt.get_name(), end_stmt.get_name()))
-                else:
-                    end_stmt.set_name(start_stmt.get_name())
         return content,
 
     def init(self, content):
@@ -606,48 +658,100 @@ content : tuple
 
 
 class SequenceBase(Base):
-    """
-::
-    <sequence-base> = <obj>, <obj> [ , <obj> ]...
-    """
+    '''
+    Match one or more fparser2 rules separated by a defined separator.
+
+    sequence-base is obj [sep obj ] ...
+
+    '''
+    @staticmethod
     def match(separator, subcls, string):
+        '''Match one or more 'subcls' fparser2 rules in the string 'string'
+        separated by 'separator'.
+
+        :param str separator: the separator used to split the supplied \
+        string.
+        :param subcls: an fparser2 object representing the rule that \
+        should be matched.
+        :type subcls: subclass of :py:class:`fparser.two.utils.Base`
+        :param str string: the input string to match.
+
+        :returns: a tuple containing 1) the separator and 2) the \
+        matched objects in a tuple, or None if there is no match.
+        :rtype: (str, (Subclass of \
+        :py:class:`fparser.two.utils.Base`)) or NoneType
+
+        :raises InternalError: if the separator or string arguments \
+        are not the expected type.
+        :raises InternalError: if the separator is white space.
+
+        '''
+        if not isinstance(separator, (str, six.text_type)):
+            raise InternalError(
+                "SequenceBase class match method argument separator expected "
+                "to be a string but found '{0}'.".format(type(string)))
+        if not isinstance(string, (str, six.text_type)):
+            raise InternalError(
+                "SequenceBase class match method argument string expected to "
+                "be a string but found '{0}'.".format(type(string)))
+
+        if separator == ' ':
+            raise InternalError(
+                "SequenceBase class match method argument separator cannot "
+                "be white space.")
         line, repmap = string_replace_map(string)
-        if isinstance(separator, str):
-            splitted = line.split(separator)
-        else:
-            splitted = separator[1].split(line)
-            separator = separator[0]
-        if len(splitted) <= 1:
-            return
-        lst = []
-        for p in splitted:
-            lst.append(subcls(repmap(p.strip())))
+        splitted = line.split(separator)
+        if not splitted:
+            # There should be at least one entry.
+            return None
+        lst = [subcls(repmap(entry.strip())) for entry in splitted]
         return separator, tuple(lst)
-    match = staticmethod(match)
 
     def init(self, separator, items):
+        '''Store the result of the match method if the match is successful.
+
+        :param str separator: The separator used to split the supplied \
+        string.
+        :param items: A tuple containing the matched objects in a \
+        tuple.
+        :type items: (Subclass of :py:class:`fparser.two.utils.Base`)
+
+        '''
         self.separator = separator
         self.items = items
-        return
 
     def tostr(self):
-        s = self.separator
-        if s == ',':
-            s = s + ' '
-        elif s == ' ':
+        '''
+        :returns: The Fortran representation of this object as a string.
+        :rtype: str
+
+        '''
+        sep = self.separator
+        if sep == ',':
+            sep = sep + ' '
+        elif sep == ' ':
             pass
         else:
-            s = ' ' + s + ' '
-        return s.join(map(str, self.items))
+            sep = ' ' + sep + ' '
+        return sep.join(map(str, self.items))
 
     def torepr(self):
-        return '%s(%r, %r)' % (self.__class__.__name__,
-                               self.separator, self.items)
+        '''
+        :returns: The Python representation of this object as a string.
+        :rtype: str
 
-    def _cmpkey(self):
-        """ Provides a key of objects to be used for comparing.
-        """
-        return (self.separator, self.items)
+        '''
+        return "{0}('{1}', {2})".format(self.__class__.__name__,
+                                        self.separator, self.items)
+
+    # The mixin class is likely to be removed so _cmpkey would not be
+    # needed. It is not used at the moment. It is only commented out
+    # at this point, rather than removed, in case it turns out that
+    # the mixin class is useful.
+    # def _cmpkey(self):
+    #     """ Provides a key of objects to be used for comparing.
+    #     """
+    #     return (self.separator, self.items)
 
 
 class UnaryOpBase(Base):
@@ -823,59 +927,92 @@ class KeywordValueBase(Base):
 
 
 class BracketBase(Base):
-    """
-::
-    <bracket-base> = <left-bracket-base> <something> <right-bracket>
-    """
+    '''
+    bracket-base is left-bracket something right-bracket.
+
+    This class is able to cope with nested brackets as long as they
+    are correctly nested. Brackets in strings are ignored.
+
+    The 'something' can be specified as being optional.
+
+    '''
     @staticmethod
     def match(brackets, cls, string, require_cls=True):
-        ''' The generic match method for all types of bracketed
-        expressions '''
-        bracket_len = len(brackets)//2
-        left = brackets[:bracket_len]
-        right = brackets[-bracket_len:]
+        '''A generic match method for all types of bracketed
+        expressions.
 
-        if not (string.startswith(left) and string.endswith(right)):
-            return
+        :param str brackets: the format of the left and right brackets \
+        provided as a string, for example '()'
+        :param cls: the class to match the content within the brackets \
+        :type cls: subclass of :py:class:`fparser.two.utils.Base`
+        :param str string: the content to match
+        :param bool require_cls: whether the class and associated \
+        content is mandatory (True) or optional (False). The default \
+        is True.
+        :return: None if there is no match, otherwise a tuple with the \
+        first and third entries being strings containing the left and \
+        right brackets respectively and the second entry being either \
+        None or an instance of the class provided as the second \
+        argument (cls).
+        :rtype: 'NoneType', ( `str`, `NoneType`, `str`) or ( `str`, \
+        `cls`, `str` )
 
+        '''
+        if not cls and require_cls:
+            return None
+        if not string:
+            return None
+        string_strip = string.strip()
+        if not brackets:
+            return None
+        brackets_nospc = brackets.replace(' ', '')
+        if not brackets_nospc:
+            return None
+        if len(brackets_nospc) % 2 == 1:
+            # LHS and RHS bracketing must be the same size
+            return None
+        bracket_len = len(brackets_nospc)//2
+        left = brackets_nospc[:bracket_len]
+        right = brackets_nospc[-bracket_len:]
+        if len(string_strip) < bracket_len*2:
+            return None
+        if not (string_strip.startswith(left) and
+                string_strip.endswith(right)):
+            return None
         # Check whether or not there's anything between the open
         # and close brackets
-        line = string[bracket_len:-bracket_len].strip()
-        if not line:
-            if require_cls:
-                return
+        line = string_strip[bracket_len:-bracket_len].strip()
+        if (not line and cls and require_cls) or (line and not cls):
+            return None
+        if not line and (not cls or not require_cls):
             return left, None, right
-
-        # There's some content between the open and close brackets.
-        # We may have something like "(a + b)*(a - b)" so have
-        # to check - we start with one open bracket. If we reach
-        # zero open brackets before we get to the end then the
-        # opening bracket at the start of the string does not
-        # correspond to the closing bracket at the end of it.
-        # Unless of course any interim brackets we encounter are
-        # within strings...
-        num_open = 1
-        in_string = False
-        for idx in range(bracket_len, len(string)-bracket_len):
-            if string[idx] == '"' or string[idx] == "'":
-                in_string = not in_string
-            if in_string:
-                # Ignore anything within quotes
-                continue
-            # A slice in python goes up to but *does not
-            # include* the last position so no need for a '-1'
-            if string[idx:idx+bracket_len] == left:
-                num_open += 1
-            elif string[idx:idx+bracket_len] == right:
-                num_open -= 1
-            if num_open == 0:
-                return
         return left, cls(line), right
 
     def tostr(self):
+        '''
+        :raises InternalError: if the internal items list variable is \
+        not the expected size.
+        :raises InternalError: if the first element of the internal \
+        items list is None or is an empty string.
+        '''
+
+        if len(self.items) != 3:
+            raise InternalError(
+                "Class BracketBase method tostr() has '{0}' items, "
+                "but expecting 3.".format(len(self.items)))
+        if not self.items[0]:
+            raise InternalError(
+                "Class BracketBase method tostr(). 'Items' entry 0 "
+                "should be a string containing the left hand bracket "
+                "but it is empty or None")
+        if not self.items[2]:
+            raise InternalError(
+                "Class BracketBase method tostr(). 'Items' entry 2 "
+                "should be a string containing the right hand bracket "
+                "but it is empty or None")
         if self.items[1] is None:
-            return '%s%s' % (self.items[0], self.items[2])
-        return '%s%s%s' % tuple(self.items)
+            return "{0}{1}".format(self.items[0], self.items[2])
+        return "{0}{1}{2}".format(self.items[0], self.items[1], self.items[2])
 
 
 class NumberBase(Base):
@@ -1002,28 +1139,68 @@ string
 
 
 class STRINGBase(StringBase):
-    """
-::
-    <STRING-base> = <XYZ>
-    """
-    match = staticmethod(StringBase.match)
+    '''STRINGBase matches an upper case version of the input string with
+    another a pattern (typically taken from pattern_tools.py) and
+    returns the string in upper case if there is a match.
 
-    def match(pattern, string):
-        if isinstance(pattern, (list, tuple)):
-            for p in pattern:
-                obj = STRINGBase.match(p, string)
-                if obj is not None:
-                    return obj
-            return
-        STRING = string.upper()
-        if isinstance(pattern, str):
-            if len(pattern) == len(string) and pattern == STRING:
-                return STRING,
-            return
-        if pattern.match(STRING):
-            return STRING,
-        return
-    match = staticmethod(match)
+    '''
+
+    @staticmethod
+    def match(my_pattern, string):
+        '''Matches an input string with a specified pattern. Casts the string
+        to upper case before performing a match and, if there is a
+        match, returns the string in upper case.
+
+        The pattern can be a regular expression, a string, a list or a
+        tuple. If the input pattern is a regular expression or a
+        string, a direct equivalence is performed. If the input pattern is a
+        list or a tuple, then all of the contents of the list
+        or tuple are searched for a match (by recursing). The list or tuple may
+        contain regular expressions, strings, lists or tuples. This
+        functionality can be used to recurse down a tree of lists and
+        or tuples until regular expressions or strings are found (at
+        the leaves of the tree) on which to match. The patterns used
+        to match in fparser can be found in patterns_tools.py. These
+        make use of the pattern class, whose match method behaves like
+        a regular expression. For example:
+
+        from fparser.two import pattern_tools
+        pattern = pattern_tools.intrinsic_type_name
+        result = STRINGBase.match(pattern, "logical")
+
+        :param pattern: the pattern to match
+        :type pattern: `list`, `tuple`, `str` or an `re` expression
+        :param str string: the string to match with the pattern
+        :return: None if there is no match, or a tuple containing the \
+        matched string in upper case.
+        :rtype: `NoneType` or ( `str` )
+
+        '''
+        if string is None:
+            return None
+        if not isinstance(string, (str, six.text_type)):
+            raise InternalError(
+                "Supplied string should be of type str or {0}, but found "
+                "{1}".format(six.text_type, type(string)))
+        if isinstance(my_pattern, (list, tuple)):
+            for child in my_pattern:
+                result = STRINGBase.match(child, string)
+                if result:
+                    return result
+            return None
+        string_upper = string.upper()
+        if isinstance(my_pattern, str):
+            if len(my_pattern) == len(string) and my_pattern == string_upper:
+                return string_upper,
+            return None
+        try:
+            if my_pattern.match(string_upper):
+                return string_upper,
+        except AttributeError:
+            raise InternalError(
+                "Supplied pattern should be a list, tuple, str or regular "
+                "expression but found {0}".format(type(my_pattern)))
+        return None
 
 
 class StmtBase(Base):
@@ -1082,7 +1259,7 @@ class EndStmtBase(StmtBase):
         else:
             if require_stmt_type:
                 return
-            line = ''
+            return None, None
         if line:
             if stmt_name is None:
                 return
@@ -1091,7 +1268,6 @@ class EndStmtBase(StmtBase):
 
     def init(self, stmt_type, stmt_name):
         self.items = [stmt_type, stmt_name]
-        self.type, self.name = stmt_type, stmt_name
         return
 
     def get_name(self):
@@ -1100,24 +1276,16 @@ class EndStmtBase(StmtBase):
     def get_type(self):
         return self.items[0]
 
-    def set_name(self, name):
-        from fparser.two.Fortran2003 import Name
-        if self.items[1] is not None:
-            self.warning(
-                'item already has name %r, changing it to %r' %
-                (self.items[1], name))
-        if isinstance(name, Name):
-            self.items[1] = name
-        else:
-            self.items[1] = Name(name)
-
     def tostr(self):
         if self.items[1] is not None:
             return 'END %s %s' % tuple(self.items)
-        return 'END %s' % (self.items[0])
+        if self.items[0] is not None:
+            return 'END %s' % (self.items[0])
+        return 'END'
 
     def torepr(self):
-        return '%s(%r, %r)' % (self.__class__.__name__, self.type, self.name)
+        return '%s(%r, %r)' % (
+            self.__class__.__name__, self.get_type(), self.get_name())
 
     def get_end_name(self):
         name = self.items[1]
